# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal exceptions that can be generated in network interface code.
These are use to control what the upper levels should do when different
types of errors occur.

"""


class NetworkInterfaceException(Exception): pass
class ForceAgentRestart(NetworkInterfaceException): pass
class ForceAgentDisconnect(NetworkInterfaceException): pass
class DiscardDataForRequest(NetworkInterfaceException): pass
class RetryDataForRequest(NetworkInterfaceException): pass
